from abc import ABC, abstractmethod
import socket
import struct
from unittest.mock import patch
from pathlib import Path

from ekfsm.devices.gpio import EKFIdSimGpio
from ekfsm.devices.gpio import SimGpio
from .core.sysfs import set_sysfs_root
from .core.components import SystemComponent

from .devices import GPIO
from typing import List
from smbus2 import SMBus

GPIO_SIM_MAPPING = {}
SMBUS_SIM_MAPPING = {}


class SimSmbus(ABC):
    @abstractmethod
    def read_word_data(self, cmd: int) -> int:
        pass

    @abstractmethod
    def read_block_data(self, cmd: int) -> List[int]:
        pass

    @abstractmethod
    def write_block_data(self, cmd: int, data: List[int]):
        pass

    @abstractmethod
    def write_byte(self, cmd: int):
        pass

    @abstractmethod
    def write_word_data(self, cmd: int, data: int):
        pass


def register_gpio_sim(major: int, minor: int, sim_gpio: SimGpio) -> None:
    name = f"{major}:{minor}"
    if name in GPIO_SIM_MAPPING:
        raise ValueError(f"GPIO_SIM_MAPPING already contains {name}")
    GPIO_SIM_MAPPING[name] = sim_gpio


def find_gpio_dev_with_major_minor(major: int, minor: int) -> SimGpio:
    name = f"{major}:{minor}"
    if name not in GPIO_SIM_MAPPING:
        raise ValueError(f"GPIO_SIM_MAPPING does not contain {name}")
    return GPIO_SIM_MAPPING[name]


def register_smbus_sim(bus_num: int, i2c_addr: int, sim_smbus: SimSmbus) -> None:
    name = f"{bus_num}:{i2c_addr}"
    if name in SMBUS_SIM_MAPPING:
        raise ValueError(f"SMBUS_SIM_MAPPING already contains {name}")
    SMBUS_SIM_MAPPING[name] = sim_smbus


def find_smbus_dev(bus_num: int, i2c_addr: int) -> SimSmbus:
    name = f"{bus_num}:{i2c_addr}"

    if name not in SMBUS_SIM_MAPPING:
        raise ValueError(f"SMBUS_SIM_MAPPING does not contain {name}")
    return SMBUS_SIM_MAPPING[name]


class GpioSimulator(GPIO):
    def __init__(
        self,
        name: str,
        parent: SystemComponent | None = None,
        *args,
        **kwargs,
    ):
        super(GPIO, self).__init__(
            name,
            parent,
            *args,
            **kwargs,
        )
        major, minor = self._find_gpio_dev(parent, *args, **kwargs)
        self._sim_gpio = find_gpio_dev_with_major_minor(major, minor)
        self.number = minor

    def num_lines(self) -> int:
        return self._sim_gpio.num_lines()

    def set_pin(self, pin: int, value: bool) -> None:
        self._sim_gpio.set_pin(pin, value)

    def get_pin(self, pin: int) -> bool:
        return self._sim_gpio.get_pin(pin)

    def set_direction(self, pin: int, direction: bool) -> None:
        self._sim_gpio.set_direction(pin, direction)

    def __str__(self) -> str:
        return f"GPIO_SIM({self.name})"


class SmbusSimulator:
    def __init__(self, bus_num: int):
        self.bus_num = bus_num

    def read_word_data(self, i2c_addr: int, cmd: int) -> int:
        return find_smbus_dev(self.bus_num, i2c_addr).read_word_data(cmd)

    def read_block_data(self, i2c_addr: int, cmd: int) -> List[int]:
        return find_smbus_dev(self.bus_num, i2c_addr).read_block_data(cmd)

    def write_block_data(self, i2c_addr: int, cmd: int, data: List[int]):
        find_smbus_dev(self.bus_num, i2c_addr).write_block_data(cmd, data)

    def write_byte(self, i2c_addr: int, cmd: int):
        find_smbus_dev(self.bus_num, i2c_addr).write_byte(cmd)

    def write_word_data(self, i2c_addr: int, cmd: int, data: int):
        find_smbus_dev(self.bus_num, i2c_addr).write_word_data(cmd, data)


def enable_gpio_simulation():
    patched_methods = []

    patched_methods.append(
        patch.object(GPIO, "__init__", new_callable=lambda: GpioSimulator.__init__)
    )
    patched_methods.append(
        patch.object(GPIO, "num_lines", new_callable=lambda: GpioSimulator.num_lines)
    )
    patched_methods.append(
        patch.object(GPIO, "set_pin", new_callable=lambda: GpioSimulator.set_pin)
    )
    patched_methods.append(
        patch.object(GPIO, "get_pin", new_callable=lambda: GpioSimulator.get_pin)
    )
    patched_methods.append(
        patch.object(
            GPIO, "set_direction", new_callable=lambda: GpioSimulator.set_direction
        )
    )
    patched_methods.append(
        patch.object(GPIO, "__str__", new_callable=lambda: GpioSimulator.__str__)
    )
    for pm in patched_methods:
        pm.start()


def enable_smbus_simulation():
    patched_methods = []

    patched_methods.append(
        patch.object(SMBus, "__init__", new_callable=lambda: SmbusSimulator.__init__)
    )
    patched_methods.append(
        patch.object(
            SMBus, "read_word_data", new_callable=lambda: SmbusSimulator.read_word_data
        )
    )
    patched_methods.append(
        patch.object(
            SMBus,
            "read_block_data",
            new_callable=lambda: SmbusSimulator.read_block_data,
        )
    )
    patched_methods.append(
        patch.object(
            SMBus,
            "write_block_data",
            new_callable=lambda: SmbusSimulator.write_block_data,
        )
    )
    patched_methods.append(
        patch.object(
            SMBus, "write_byte", new_callable=lambda: SmbusSimulator.write_byte
        )
    )
    patched_methods.append(
        patch.object(
            SMBus,
            "write_word_data",
            new_callable=lambda: SmbusSimulator.write_word_data,
        )
    )

    for pm in patched_methods:
        pm.start()


def enable_simulation(sysfs_path: Path):
    global GPIO_SIM_MAPPING
    GPIO_SIM_MAPPING = {}

    global SMBUS_SIM_MAPPING
    SMBUS_SIM_MAPPING = {}

    set_sysfs_root(sysfs_path)
    enable_gpio_simulation()
    enable_smbus_simulation()


def register_gpio_simulations():
    register_gpio_sim(233, 1, EKFIdSimGpio(0x38, 0x1, 0x0, 0x6))  # SRF Rev 0
    register_gpio_sim(233, 2, EKFIdSimGpio(0x34, 0xA, 0x0, 0x1))  # CCU Rev 0


class SocketSmbus(SimSmbus):
    def __init__(self, host: str, port: int) -> None:
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((host, port))

    def read_word_data(self, cmd: int) -> int:
        req = struct.pack("BB", 0x4, cmd)
        self.sock.send(req)
        data = self.sock.recv(2)
        return struct.unpack("<H", data)[0]

    def read_block_data(self, cmd: int) -> List[int]:
        req = struct.pack("BB", 0x1, cmd)
        self.sock.send(req)
        _count = self.sock.recv(1)
        count = struct.unpack("B", _count)[0]
        data = self.sock.recv(count)
        return [int(data[i]) for i in range(0, len(data), 1)]

    def write_block_data(self, cmd: int, data: List[int]):
        _data = bytes(data)
        hdr = struct.pack("BBB", 0x2, cmd, len(_data))
        self.sock.send(hdr + _data)

    def write_byte(self, cmd: int):
        hdr = struct.pack("BB", 0x3, cmd)
        self.sock.send(hdr)

    def write_word_data(self, cmd: int, data: int):
        hdr = struct.pack("BBH", 0x5, cmd, data)
        self.sock.send(hdr)



from .slots import Slot, SlotType, Slots  # noqa: F401
from .sysfs import SysFSDevice, SysFSAttribute  # noqa: F401
from .components import HwModule  # noqa: F401
from .probe import ProbeableDevice  # noqa: F401



from typing import TYPE_CHECKING
from munch import Munch
from ekfsm.core.components import SystemComponent
from ekfsm.core.sysfs import SysFSDevice, sysfs_root
from ekfsm.exceptions import ConfigError
from pathlib import Path

if TYPE_CHECKING:
    from ekfsm.core.components import HwModule


class Device(SystemComponent):
    """
    A generic device.
    """

    def __init__(
        self,
        name: str,
        parent: SystemComponent | None = None,
        children: list["Device"] | None = None,
        *args,
        **kwargs
    ):
        super().__init__(name)
        self.parent = parent
        self.device_args = kwargs
        self.logger.debug(f"Device: {name} {kwargs}")

        if children:
            self.children = children

        if not hasattr(self, "sysfs_device"):
            self.sysfs_device: SysFSDevice | None = None

        self._provides_attrs = kwargs.get("provides", {})

        self.provides = self.__post_init__(Munch(self._provides_attrs))

    def __post_init__(self, provides: Munch) -> Munch:
        return self.__init_dynamic_attrs__(provides)

    def __init_dynamic_attrs__(self, provides: Munch) -> Munch:

        for key, fields in provides.items():
            if isinstance(fields, dict):
                provides[key] = self.__init_dynamic_attrs__(Munch(fields))
            elif isinstance(fields, list | str):
                provides[key] = Munch()

                if isinstance(fields, str):
                    fields = [fields]

                while fields:
                    iface = fields.pop()
                    if isinstance(iface, dict):
                        name = list(iface.keys())[0]
                        try:
                            func = list(iface.values())[0]
                        except IndexError:
                            raise ConfigError(
                                f"{self.name}: No function given for interface {name}."
                            )
                        if not hasattr(self, func):
                            raise NotImplementedError(
                                f"{self.name}: Function {func} for interface {name} not implemented."
                            )
                        provides[key].update({name: getattr(self, func)})
                    else:
                        if not hasattr(self, iface):
                            raise NotImplementedError(
                                f"{self.name}: Function {iface} for provider {key} not implemented."
                            )
                        provides[key].update({iface: getattr(self, iface)})

        return provides

    def read_sysfs_attr_bytes(self, attr: str) -> bytes | None:
        if self.sysfs_device and len(attr) != 0:
            return self.sysfs_device.read_attr_bytes(attr)
        return None

    def read_sysfs_attr_utf8(self, attr: str) -> str | None:
        if self.sysfs_device and len(attr) != 0:
            return self.sysfs_device.read_attr_utf8(attr)
        return None

    def write_sysfs_attr(self, attr: str, data: str | bytes) -> None:
        if self.sysfs_device and len(attr) != 0:
            return self.sysfs_device.write_attr(attr, data)
        return None

    @property
    def hw_module(self) -> 'HwModule':
        from ekfsm.core.components import HwModule

        if isinstance(self.root, HwModule):
            return self.root
        else:
            raise RuntimeError("Device is not a child of HwModule")

    def get_i2c_chip_addr(self) -> int:
        assert self.parent is not None

        chip_addr = self.device_args.get("addr")
        if chip_addr is None:
            raise ConfigError(
                f"{self.name}: Chip address not provided in board definition"
            )

        if not hasattr(self.parent, "sysfs_device") or self.parent.sysfs_device is None:
            # our device is the top level device of the slot
            # compute chip address from board yaml and slot attributes
            slot_attributes = self.hw_module.slot.attributes

            if slot_attributes is None:
                raise ConfigError(
                    f"{self.name}: Slot attributes not provided in system configuration"
                )

            if not self.hw_module.is_master:
                # slot coding is only used for non-master devices
                if not hasattr(slot_attributes, "slot_coding"):
                    raise ConfigError(
                        f"{self.name}: Slot coding not provided in slot attributes"
                    )

                slot_coding_mask = 0xFF

                if hasattr(slot_attributes, "slot_coding_mask"):
                    slot_coding_mask = slot_attributes.slot_coding_mask

                chip_addr |= slot_attributes.slot_coding & slot_coding_mask

        return chip_addr

    def get_i2c_sysfs_device(self, addr: int) -> SysFSDevice:
        from ekfsm.core.components import HwModule

        parent = self.parent
        assert parent is not None

        # if parent is a HwModule, we can get the i2c bus from the master device
        if isinstance(parent, HwModule):
            i2c_bus_path = self._master_i2c_bus()
        else:
            # otherwise the parent must be a MuxChannel
            from ekfsm.devices.mux import MuxChannel

            assert isinstance(parent, MuxChannel)
            assert parent.sysfs_device is not None
            i2c_bus_path = parent.sysfs_device.path

        # search for device with addr
        for entry in i2c_bus_path.iterdir():
            if (
                entry.is_dir()
                and not (entry / "new_device").exists()  # skip bus entries
                and (entry / "name").exists()
            ):
                # for PRP devices, address is contained in firmware_node/description
                if (entry / "firmware_node").exists() and (
                    entry / "firmware_node" / "description"
                ).exists():
                    description = (
                        (entry / "firmware_node/description").read_text().strip()
                    )
                    got_addr = int(description.split(" - ")[0], 16)
                    if got_addr == addr:
                        return SysFSDevice(entry)

                # for non-PRP devices, address is contained in the directory name (e.g. 2-0018)
                else:
                    got_addr = int(entry.name.split("-")[1], 16)
                    if got_addr == addr:
                        return SysFSDevice(entry)

        raise FileNotFoundError(
            f"Device with address 0x{addr:x} not found in {i2c_bus_path}"
        )

    @staticmethod
    def _master_i2c_get_config(master: "HwModule") -> dict:
        if (
            master.config.get("bus_masters") is not None
            and master.config["bus_masters"].get("i2c") is not None
        ):
            return master.config["bus_masters"]["i2c"]
        else:
            raise ConfigError("Master definition incomplete")

    def _master_i2c_bus(self) -> Path:
        if self.hw_module.is_master:
            # we are the master
            master = self.hw_module
            master_key = "MASTER_LOCAL_DEFAULT"
            override_master_key = self.device_args.get("i2c_master", None)
            if override_master_key is not None:
                master_key = override_master_key
        else:
            # another board is the master
            if self.hw_module.slot.master is None:
                raise ConfigError(
                    f"{self.name}: Master board not found in slot attributes"
                )

            master = self.hw_module.slot.master
            master_key = self.hw_module.slot.slot_type.name

        i2c_masters = self._master_i2c_get_config(master)

        if i2c_masters.get(master_key) is not None:
            dir = sysfs_root() / Path(i2c_masters[master_key])
            bus_dirs = list(dir.glob("i2c-*"))
            if len(bus_dirs) == 1:
                return bus_dirs[0]
            elif len(bus_dirs) > 1:
                raise ConfigError(f"Multiple master i2c buses found for {master_key}")
            raise ConfigError(f"No master i2c bus found for {master_key}")
        else:
            raise ConfigError(f"Master i2c bus not found for {master_key}")

    def get_i2c_bus_number(self) -> int:
        """
        Get the I2C bus number of the device. Works for devices that do not have a sysfs_device attribute.
        """
        from ekfsm.devices.mux import MuxChannel

        if isinstance(self, MuxChannel):
            raise RuntimeError(f"{self.name}: MuxChannel does not have a bus number")

        if self.sysfs_device is None:
            if self.parent is None:
                raise RuntimeError(f"{self.name}: Must have a parent to get bus number")
            parent_path = self.parent.sysfs_device.path
        else:
            parent_path = self.sysfs_device.path.parent
        if parent_path.is_symlink():
            parent_path = parent_path.readlink()
        bus_number = parent_path.name.split("-")[1]
        return int(bus_number)

    def __repr__(self) -> str:
        sysfs_path = getattr(self.sysfs_device, "path", "")
        return f"{self.name}; sysfs_path: {sysfs_path}"



from pathlib import Path
from ekfsm.system import System
import logging

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

config = Path(__file__).parent / "zip2-smoke.yaml"
system = System(config)


cpu = system[0]
print(f"probing CPU: {cpu.probe()}")
print(
    f"inventory: {cpu.inventory.vendor()} {cpu.inventory.model()} {cpu.inventory.serial()}"
)

sq1 = system[1]
print(f"probing SQ1: {sq1.probe()}")
print(
    f"inventory: {sq1.inventory.vendor()} {sq1.inventory.model()} {sq1.inventory.serial()}"
)

sur = system.sur
print(f"probing SUR: {sur.probe()}")
print(
    f"inventory: {sur.inventory.vendor()} {sur.inventory.model()} {sur.inventory.serial()}"
)

spv = system.spv
print(f"probing SPV: {spv.probe()}")
print(
    f"inventory: {spv.inventory.vendor()} {spv.inventory.model()} {spv.inventory.serial()}"
)

print(system)

# SUR LED Test
system.sur.led_a.set(0, "red")
system.sur.led_a.set(1, "yellow")
system.sur.led_c.set(0, "white")
system.sur.led_d.set(1, "blue")

print(f"SUR TEMP: {system.sur.th.temperature()}")
print(f"SUR HUMIDITY: {system.sur.th.humidity()}")


# CCU Test
ccu = system.ccu
print(f"CCU firmware: {ccu.management.identify_firmware()}")

# SPV Test
print(f"modem1 control lines {spv.gpio0.num_lines()}")
print(f"modem1 control lines {spv.gpio1.num_lines()}")
print(f"modem1 control lines {spv.gpio2.num_lines()}")



from ekfsm.core.components import SystemComponent

from ..core.sysfs import SysFSDevice
from .generic import Device


class MuxChannel(Device):
    def __init__(
        self,
        name: str,
        channel_id: int,
        parent: 'I2CMux',
        children: list[Device] | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(name=name, parent=parent, children=children)
        self.channel_id = channel_id

        assert parent.sysfs_device is not None
        assert isinstance(self.parent, I2CMux)

        path = parent.sysfs_device.path / f"channel-{self.channel_id}"
        self.sysfs_device = SysFSDevice(path)


class I2CMux(Device):
    def __init__(
        self,
        name: str,
        parent: SystemComponent | None = None,
        children: list[MuxChannel] | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, children, **kwargs)

        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)



from pathlib import Path

from ekfsm.core.components import SystemComponent

from ..core.sysfs import SysFSDevice

from .generic import Device
from ..core.probe import ProbeableDevice


class PmBus(Device, ProbeableDevice):
    def __init__(
        self,
        name: str,
        parent: SystemComponent | None = None,
        children: list[Device] | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(name, parent, children, **kwargs)
        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)

        files = list(Path(self.sysfs_device.path).rglob("hwmon/*/in1_input"))
        if len(files) == 0:
            raise FileNotFoundError("No HWMON entries found")
        self.hwmon_sysfs = SysFSDevice(files[0].parent)

    def probe(self, *args, **kwargs) -> bool:
        from ekfsm.core import HwModule

        assert isinstance(self.root, HwModule)
        return self.root.id == self.model()

    # Voltage and Current Interfaces
    def _in_conversion(self, in_file: str) -> float:
        return float(self.hwmon_sysfs.read_attr_utf8(in_file)) / 1000.0

    def _current_conversion(self, in_file: str) -> float:
        return float(self.hwmon_sysfs.read_attr_utf8(in_file)) / 1000.0

    def in1_input(self) -> float:
        return self._in_conversion("in1_input")

    def in2_input(self) -> float:
        return self._in_conversion("in2_input")

    def curr1_input(self) -> float:
        return self._current_conversion("curr1_input")

    def curr2_input(self) -> float:
        return self._current_conversion("curr2_input")

    # Inventory Interface
    def vendor(self) -> str:
        return self.hwmon_sysfs.read_attr_utf8("vendor").strip()

    def model(self) -> str:
        return self.hwmon_sysfs.read_attr_utf8("model").strip()

    def serial(self) -> str:
        return self.hwmon_sysfs.read_attr_utf8("serial").strip()

    def revision(self) -> str:
        return self.hwmon_sysfs.read_attr_utf8("revision").strip()



from ekfsm.devices.generic import Device
from ekfsm.devices.hwmon import HWMON
from ekfsm.devices.smbios import SMBIOS
from .eeprom import EEPROM, EKF_EEPROM, EKF_CCU_EEPROM
from .pmbus import PmBus
from .gpio import GPIO, EKFIdentificationIOExpander, GPIOExpander
from .ekf_sur_led import EKFSurLed
from .ekf_ccu_uc import EKFCcuUc
from .iio_thermal_humidity import IIOThermalHumidity
from .mux import I2CMux, MuxChannel

CLASS_MAP = {
    "GenericDevice": Device,
    "I2CMux": I2CMux,
    "MuxChannel": MuxChannel,
    "GPIO": GPIO,
    "GPIOExpander": GPIOExpander,
    "EKFIdentificationIOExpander": EKFIdentificationIOExpander,
    "EEPROM": EEPROM,
    "EKF_EEPROM": EKF_EEPROM,
    "EKF_CCU_EEPROM": EKF_CCU_EEPROM,
    "EKFCcuUc": EKFCcuUc,
    "PmBus": PmBus,
    "SMBIOS": SMBIOS,
    "HWMON": HWMON,
    "EKFSurLed": EKFSurLed,
    "IIOThermalHumidity": IIOThermalHumidity,
}



from __future__ import annotations
import logging
from typing import TYPE_CHECKING
from anytree import AnyNode

from ekfsm.devices import CLASS_MAP
from ekfsm.exceptions import ConfigError

if TYPE_CHECKING:
    from .components import HwModule
    from ekfsm.devices.generic import Device


class BoardDictImporter:

    def __init__(self, nodecls=AnyNode):
        self.nodecls = nodecls

    def import_(self, logger: logging.Logger, data, parent=None):
        """Import tree from `data`."""
        return self.__import(logger, data, parent=parent)

    def __import(self, logger: logging.Logger, data, parent=None):
        from .components import HwModule

        device_type = data.get("device_type")
        nodecls = CLASS_MAP.get(device_type)
        if nodecls is None:
            raise ConfigError(f"Unknown device type: {device_type}")

        children = data.pop("children", [])
        if parent is not None and isinstance(parent, HwModule):
            pass

        node = nodecls(parent=parent, **data)

        if children is not None:
            for child in children:
                try:
                    logger.debug(f"Importing sub device {child}")
                    self.__import(logger, child, parent=node)
                except Exception as e:
                    logger.error(
                        f"Failed to import sub device {child}: {e}. continue anyway"
                    )
        return node


def deserialize_hardware_tree(
    logger: logging.Logger, data: dict, parent: "HwModule"
) -> tuple[str, str, str, list["Device"]]:
    importer = BoardDictImporter()

    id = data.pop("id", None)
    if id is None:
        raise ConfigError("Board configuration must contain `id`")
    name = data.pop("name", None)
    if name is None:
        raise ConfigError("Board configuration must contain `name`")
    slot_type = data.pop("slot_type")
    if slot_type is None:
        raise ConfigError("Board configuration must contain `board_type`")

    children = data.pop("children", None)
    devices = []
    if children:
        for child in children:
            try:
                logger.debug(f"Importing top level device {child}")
                node = importer.import_(logger, child, parent=parent)
                devices.append(node)
            except Exception as e:
                logger.error(
                    f"Failed to import top level device {child}: {e}. continue anyway"
                )

    return id, name, slot_type, devices



from .core.slots import Slot
from .system import System
from .core import HwModule
from .core.slots import SlotType
from .exceptions import ConfigError

__all__ = (
    "System",
    "ConfigError",
    "HwModule",
    "Slot",
    "SlotType",
)



import yamale  # type: ignore
import yaml
import munch

from typing import Any, List, Tuple

schema_str = """
system_config:
  name: str()
  slots: list(include('slot'))
---
slot:
  name: str()
  slot_type: regex('^CPCI_S0_(PER|SYS|PSU|UTILITY)$')
  desired_hwmodule_type: str()
  desired_hwmodule_name: str()
  attributes: map(required=False)
"""


def _validate_config(config_file: str) -> None:
    schema = yamale.make_schema(content=schema_str)
    data = yamale.make_data(config_file)
    yamale.validate(schema, data)


def _parse_config(config_file: str) -> Any | munch.Munch | List | Tuple:
    with open(config_file) as file:
        config = yaml.safe_load(file)
        munchified_config = munch.munchify(config)
    return munchified_config


def load_config(config_file: str) -> Any:
    _validate_config(config_file)
    config = _parse_config(config_file)
    return config



from pathlib import Path

SYSFS_ROOT = Path("/sys")


def sysfs_root() -> Path:
    return SYSFS_ROOT


def set_sysfs_root(path: Path) -> None:
    global SYSFS_ROOT
    SYSFS_ROOT = path


def file_is_sysfs_attr(path: Path) -> bool:
    return path.is_file() and not path.stat().st_mode & 0o111


class SysFSAttribute:
    """
    A SysFSAttribute is a singular sysfs attribute located somewhere in */sys*.

    Attributes
    ----------
    path : Path
        path to the underlying file for the SysFSAttribute instance
    name : str
        attribute name, or the last part of self.path

    Methods
    -------
    read() -> Optional[str]
        Reads the file attribute if it exists
    write(data: Union[str, int]) -> None
        Writes data to the file attribute

    """

    def __init__(self, path: Path):
        self.path = path
        if not path.exists() or not path.is_file():
            raise FileNotFoundError("Invalid sysfs attribute path")

        self.name: str = path.name

    def read_utf8(self) -> str:
        return self.path.read_text()

    def read_bytes(self) -> bytes:
        return self.path.read_bytes()

    def write(self, data: str | bytes | None, offset: int = 0) -> None:
        if self.is_sysfs_attr() and data is not None:
            mode = 'r+' if isinstance(data, str) else 'rb+'
            with open(self.path, mode) as f:
                f.seek(offset)
                f.write(data)

    def is_sysfs_attr(self) -> bool:
        return file_is_sysfs_attr(self.path)

    def __repr__(self):
        return f"SysFSAttribute: {self.name}"


def list_sysfs_attributes(path: Path) -> list[SysFSAttribute]:
    if not path.exists() or not path.is_dir():
        raise FileNotFoundError(f"Invalid sysfs directory: {path}")

    return [SysFSAttribute(item) for item in path.iterdir() if file_is_sysfs_attr(item)]


class SysFSDevice:
    def __init__(self, base_dir: Path):
        self.path: Path = base_dir
        self.attributes: list[SysFSAttribute] = list_sysfs_attributes(self.path)

    def pre(self) -> None:
        pass

    def post(self) -> None:
        pass

    def write_attr(self, attr: str, data: str | bytes, offset: int = 0) -> None:
        next(x for x in self.attributes if x.name == attr).write(data, offset)

    def read_attr_utf8(self, attr: str) -> str:
        return next(x for x in self.attributes if x.name == attr).read_utf8()

    def read_attr_bytes(self, attr: str) -> bytes:
        return next(x for x in self.attributes if x.name == attr).read_bytes()



from __future__ import annotations
import os
import glob
from pathlib import Path

from ekfsm.core.sysfs import SYSFS_ROOT
from ekfsm.devices.generic import Device

# Path to the root of the HWMON sysfs filesystem
HWMON_ROOT = SYSFS_ROOT / Path('class/hwmon')


def find_core_temp_dir(hwmon_dir) -> Path:
    """
    Find the directory containing the coretemp hwmon device.

    Args:
        hwmon_dir: Path to the hwmon directory

    Returns:
        Path to the directory containing the coretemp hwmon device

    Raises:
        FileNotFoundError: If no coretemp directory is found
    """
    # List all 'name' files in each subdirectory of hwmon_dir
    name_files = glob.glob(os.path.join(hwmon_dir, '*', 'name'))

    # Search for the file containing "coretemp"
    for name_file in name_files:
        with open(name_file, 'r') as file:
            if file.readline().strip() == 'coretemp':
                # Return the directory containing this file
                return Path(os.path.dirname(name_file))

    raise FileNotFoundError("No coretemp directory found")


class HWMON(Device):
    """
    A class to represent the HWMON device.

    A HWMON device is a virtual device that is used to read hardware monitoring values from the sysfs filesystem.

    Note:
    Currently, only the CPU temperature is read from the HWMON device.
    """

    def __init__(
        self,
        name: str,
        parent: Device,
        *args,
        **kwargs,
    ):
        from ekfsm.core.sysfs import SysFSDevice, SYSFS_ROOT

        self.sysfs_device: SysFSDevice = SysFSDevice(find_core_temp_dir(SYSFS_ROOT / Path('class/hwmon')))

        super().__init__(name, parent, None, *args, **kwargs)

    def cputemp(self):
        """
        Get the CPU temperature from the HWMON device.

        Returns
        -------
        int
            The CPU temperature in degrees Celsius.
        """
        return int(self.sysfs_device.read_attr_utf8("temp1_input").strip()) / 1000



from .generic import SysFSDevice


def iio_get_in_value(dev: SysFSDevice, attrset: str) -> float:
    """
    Calculate a value from an IIO in_* attribute set, using the _raw, _scale and _offset attributes (if present).
    Typical name of attrset are "in_temp" or "in_voltage0".

    Formula according to https://wiki.st.com/stm32mpu/wiki/How_to_use_the_IIO_user_space_interface

    Parameters
    ----------
    dev
        sysfs device object pointing to the iio directory
    attrset
        name of the attribute set to read from (e.g. "in_temp")

    Returns
    -------
    float
        calculated value from the attribute set (no unit conversion)

    Raises
    ------
    FileNotFoundError
        if the neiter _input nor _raw attribute is found

    """
    # check if _input exists
    try:
        content = dev.read_attr_utf8(f"{attrset}_input")
        return float(content)
    except StopIteration:
        pass

    # check if _raw exists
    try:
        content = dev.read_attr_utf8(f"{attrset}_raw")
    except StopIteration:
        raise FileNotFoundError(f"{attrset}_raw not found")

    raw = float(content)

    # use offset if present
    try:
        content = dev.read_attr_utf8(f"{attrset}_offset")
        raw += float(content)
    except StopIteration:
        pass

    # use scale if present
    try:
        content = dev.read_attr_utf8(f"{attrset}_scale")
        raw *= float(content)
    except StopIteration:
        pass

    return raw



#! /usr/bin/env python3
# import os
# import sys
import click
import logging
from pathlib import Path

from ekfsm.system import System
from ekfsm.log import ekfsm_logger
from ekfsm.simctrl import enable_simulation
from ekfsm.simctrl import register_gpio_simulations


logging.basicConfig(level=logging.WARNING, format="%(levelname)s: %(message)s")
logger = ekfsm_logger(__name__)

sm: System | None = None


@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
@click.option("--debug", "-d", is_flag=True, help="Enable debug output")
@click.option(
    "--sysfs",
    "-s",
    help="Use custom sysfs dir for simulation mode",
    is_flag=False,
    flag_value="tests/sim/sys",
    type=click.Path(exists=True),
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="Path to configuration file",
)
def cli(verbose, debug, sysfs, config):
    global sm
    """POSIX-compliant CLI tool with subcommands"""
    if verbose:
        logging.getLogger().setLevel(logging.INFO)
        logger.info("Verbose output enabled")
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.debug("Debug output enabled")
    if sysfs:
        logger.warning("Simulation mode enabled")
        enable_simulation(Path(sysfs))
        register_gpio_simulations()

    sm = System(config)


@cli.command()
@click.option(
    "--serial",
    "-s",
    is_flag=False,
    prompt=True,
    type=int,
    help="Write chassis serial number",
)
@click.option(
    "--unit", "-u", is_flag=False, prompt=True, type=int, help="Write chassis unit"
)
@click.option(
    "--vendor",
    "-n",
    is_flag=False,
    help="Write chassis vendor",
    default="EKF Elektronik",
)
@click.option("--revision", "-r", is_flag=False, help="Write chassis revision")
@click.option("--model", "-m", is_flag=False, help="Write chassis model")
@click.option("--custom", "-c", is_flag=False, help="Write chassis custom information")
def write(serial, unit, vendor, revision, model, custom):
    """Write data to the system"""
    chassis = sm.ccu.chassis_inventory
    if serial:
        chassis.write_serial(serial)
    if unit:
        chassis.write_unit(unit)
    if vendor:
        chassis.write_vendor(vendor)
    if revision:
        chassis.write_revision(revision)
    if model:
        chassis.write_model(model)
    if custom:
        chassis.write_customer_area(custom)


@cli.command()
def show():
    """Show information about the system"""
    sm.print()


# Attach subgroups to main CLI
cli.add_command(write)
cli.add_command(show)


def main():
    cli(auto_envvar_prefix="EKFSM")


if __name__ == "__main__":
    main()



from crcmod.predefined import Crc
from typing import Sequence


def compute_int_from_bytes(data: Sequence[int]) -> int:
    # Combine the bytes into a single integer
    result = 0
    for num in data:
        result = (result << 8) | num
    return result


def get_crc16_xmodem(data: bytes) -> int:
    crc16_xmodem = Crc("xmodem")
    crc16_xmodem.update(data)
    return crc16_xmodem.crcValue



from pathlib import Path

from ekfsm.core.components import SystemComponent
from ekfsm.log import ekfsm_logger

from ..core.sysfs import SysFSDevice

from .generic import Device
from .iio import iio_get_in_value


class IIOThermalHumidity(Device):
    """
    Device for IIO thermal and/or humidity sensors.

    """

    def __init__(
        self,
        name: str,
        parent: SystemComponent | None = None,
        children: list[Device] | None = None,
        *args,
        **kwargs,
    ):
        self.logger = ekfsm_logger("IIOThermalHumidity:" + name)
        super().__init__(name, parent, children, **kwargs)
        self.addr = self.get_i2c_chip_addr()
        self.sysfs_device = self.get_i2c_sysfs_device(self.addr)

        dir = list(Path(self.sysfs_device.path).glob("iio:device*"))
        if len(dir) == 0:
            raise FileNotFoundError("iio entry not found")
        self.iio_sysfs = SysFSDevice(dir[0])
        self.logger.debug(f"iio: {self.iio_sysfs.path}")

    def temperature(self) -> float:
        return iio_get_in_value(self.iio_sysfs, "in_temp") / 1000.0

    def humidity(self) -> float:
        return iio_get_in_value(self.iio_sysfs, "in_humidityrelative") / 1000.0
